/********************************************************************/
/*                                                                  */
/*   Program Name :  speana                                         */
/*                                                                  */
/*      Version:  2.14  2007-11-04                                  */
/*                                                                  */
/*      Copyright (c) 2002-2007 T.Kondo/NICT All Right Reserved     */
/*                                                                  */
/********************************************************************/
/*
D:\IPVLBI\data\k5vssp32\kas11ft32a.dat 21 1 0 K5/VSSP32 2
  Spectrum analysis of IP-VLBI board data
        (off-line processing)

          T.Kondo 2002-05-20
               Ver 0.00 2002-05-20 
               Ver 1.41 2002-10-01 18000->16000 bug fix
               Ver 1.42 2002-10-11 set default frequency axis linear
               Ver 1.43 2004-10-11 support sampling Frequency > 16MHz
               Ver 1.44 2005-06-12 update due to update in common header name
               Ver 1.45 2005-09-27 Allow to use PGDISP as an environmetal variable
                                   for PPGPLOT device
               Ver 2.00 2005-10-28 Improve processing speed drastically
                                   support K5/VSSP32 data processing
                                   FFTW3.0 is supported if hope so.
               Ver 2.10 2006-01-31 support linear power axis
               Ver 2.11 2007-02-21 update due to libvssp.h update
			                       (Ver2.10 didn't run with new libvssp32.h)
               Ver 2.12 2007-10-26 leave window open in case of Windows
               Ver 2.13 2007-11-02 support GNUPLOT
               Ver 2.14 2007-11-04 improve appearance in case of GNUPLOT postscript

    How to run:
       speana filename [mode [sekibun [pmode [comment [soffset]]]]]

        where filename -- data file name
                            0 means "tds.data" file
               mode -- YX: axis scale mode
                          X: frequency axis mode
                             =0: log scale  =1: linear scale (default)
                          Y: amplitude axis mode
                             =0: log scale (default)  
                             =1: linear scale (common for 4 channels)
                             =2: linear scale (each max and min)
                      -value: auto correlation (value point center)
                      -1:  auto correlation (full range)
                      
               sekibun  -- integration period (integer second)
                            default is 1 
               pmode -- plot device mode
                       0: XTERM and PostScript file out (default)
                       1: PostScript out only 
                       2: XTERM only 
               comment  -- comment (displayed at upper side of graph)
                           (do not include space)
                           If omitted, you will be asked later.
                           NOTE: In the interactive mode, you can sue space.
               soffset -- start offset (sec) (default is 0)


D:\IPVLBI\data\k5vssp32\VSSP32sample.dat 0 0.01 0 TEST
*/
#define SPEANA_VERSION "Ver. 2007-11-04"

//#define USE_GNUPLOT    // use GNUPLOT  (default is use of PGPLOT)

#include "g_head.h"    // header for graphics

#ifdef WIN32
  #define USE_FFTW    // use FFTW package
#endif

#ifdef USE_FFTW
 #ifdef WIN32
  #include "fftw3.h"     // use of FFTW package
 #else
  #include <fftw3.h>     // use of FFTW package
 #endif
#endif

#include "libk5.h"      // common header for IP-VLBI utilities

#include "libfft.h"

#ifdef WIN32
#define USE_ASM    // use inline assemble function
#endif

#include "libvssp32.h"

/*    Offset from dB to dBm conversion from A/D multibit sampling data */ 
#define OFFSET1  16.0     //1 bit sampling full power convert to 0 dB (theoretical)
      /* following constants are provisional  (measured on 2002/5/18) */
#define OFFSET2   9.0     //2 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET4  -1.0     //4 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET8 -26.0     //8 bit sampling data offset measured at 200kHz for PC(vlbi11)

#ifdef WIN32
int tempcounter;  // counter for tempfile manage
#endif

#include "libspeplot.c"


int main(int argc, char *argv[])
{
    //int nloop,sekibun;
    int nloop;
    int i,imax, m, k, kaisu;
    int smode,usampl,nfft;
	int kbytes;
    int bits, bytes, numb;
    //int amode;    // >0 auto correlation plot mode 
    //int mode;  // frequency axis mode  0: log  1: linear 
    //int mode2;  // 2 digits axis mode
    int ymode; // amplitude scale mode   0: log   1: linear
    int scount[4];
    float fsampl,tsampl,fact;
    double *cdata;
    double *f_hanning;
    double theta;
    double sum,dc;
    float pows;
    float pmax,pmin,pmaxdbm,pmindbm,cofst;
    float ymaxdb,ymindb;
#ifdef USE_FFTW
    fftw_complex *cc;
    fftw_plan pfcc, pbcc;
#endif
    double nfact;
    double tinteg,tusampl;
    double telapsed;           // to measure processing time
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char ltime1st[60];
    char pgdevice_dflt[20];
    char* dflt_fname="tds.data";
    K5VSSP_file_t K5;
    speana_param_t st;

    /* set graphic device */
    strcpy(st.ps_fname,PS_OUT_FILE);
#ifdef USE_GNUPLOT   // using gnuplot
    strcpy(st.g_display,G_DISPLAY);
    strcpy(pgdevice_dflt,G_DISPLAY);
    strcpy(st.g_ps_device,G_PS_DEVICE);
#else
    strcpy(st.g_display,select_env_dflt(PGDISP, PGDISP_DFLT));
    strcpy(pgdevice_dflt,st.g_display);
    sprintf(st.g_ps_device,"%s%s\0",st.ps_fname,G_PS_DEVICE);
#endif

    if(argc == 1){
          printf("speana %s",SPEANA_VERSION);
#ifdef USE_FFTW
          printf(" compliled with FFTW3.0 \n");
#endif
#ifdef USE_GNUPLOT   /* using gnuplot */
          printf(" compliled for GNUPLOT \n");
#else
          printf(" compliled for PGPLOT \n");
#endif
          printf("\n");
          printf("Usage: speana filename [mode [sekibun [pmode [comment [soffset]]]]]\n");
          printf("\n");
          printf("         where  filename ---- Data file name\n");
          printf("                              if 0 is given, tds.data will be used\n");
          printf("                mode -------- set axis mode or autocorrelation mode\n");
          printf("                              =0 : amplitude log, frequency log\n");
          printf("                              =1 (deflt) : amplitude log, frequency linear\n");
          printf("                              =10 : amplitude linear, frequency log\n");
          printf("                              =11 : amplitude linear, frequency linear\n");
          printf("                              =-1 : Auto correlation function plot\n");
          printf("                                     instead of spectrum\n");
          printf("                              =-N : N points auto correlation function plot\n");
          printf("                sekibun --- Averaging period (integer seconds)\n");
          printf("                             default is 1\n");
          printf("                pmode ---- Plot device selection\n");
          printf("                           =0 :  \"%s\" and PostScript file (%s) out\n",pgdevice_dflt,st.ps_fname);
          printf("                           =1 :  PostScript file (%s) out only\n",st.ps_fname);
          printf("                           =2 :  \"%s\" only\n",pgdevice_dflt);
          printf("                          default is 0\n");
          printf("                comment --- Comment appeared on a graph\n");
          printf("                            if omitted, you will be asked later.\n");
          printf("                soffset --- Start time offset in sec (integer seconds)\n");
          printf("                             default is 0\n"); 
#ifndef USE_GNUPLOT   // using gnuplot
          printf("\n");
          printf("  Note: %s ---  environment variable for PGPLOT display device\n",PGDISP);
          printf("                   = %s",getenv(PGDISP));
          if(getenv(PGDISP)==NULL) {
              printf(",  program deflt is ( %s )\n",PGDISP_DFLT);
          } else {
              printf("\n");
          }
#endif
          printf("\n");
          exit(0);
    } 

    strcpy(st.fname,*(argv+1));
    if (st.fname[0]=='0'){
          strcpy(st.fname,dflt_fname);
    }
    strcpy(K5.fname,st.fname);
    /* plot mode default set */
    st.mode2=1;
    st.mode=1;   // default frequency axis mode (=Linear scale)
    ymode=0;  // default amplitude axis mode (=LOG scale)
    st.amode=0;  // auto correlation mode off
    /* plot mode set from operator */
    if (argc>=3){
       st.mode2 = atoi(*(argv + 2 ));
       if(st.mode2 <0 ){
            st.amode=-st.mode2;
            st.mode=0;
       } else {
            st.amode=0;
            ymode=st.mode2/10;
            st.mode=st.mode2 % 10;
            
       }
    }

    if (argc>=4){
       //sekibun = atoi(*(argv + 3 ));
       st.sekibun = atof(*(argv + 3 ));
    } else {
          st.sekibun=1.0;     // default integration time (sec)
    }

    if (argc>=5){
       st.pmode = atoi(*(argv + 4 ));
    } else {
          st.pmode=0;     // default plot mode
    }

    printf("speana %s",SPEANA_VERSION);
#ifdef USE_FFTW
    printf(" compliled with FFTW3.0 \n");
#else
    printf("\n");
#endif

    cstart1sec = clock();

    printf("Data File is %s\n",K5.fname);
    K5.size=get_file_size(K5.fname);
    if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",K5.fname);
        goto exitend2;
    }
    k5vssp_t_ini(&K5);

    /* get comment */

    if (argc>=6){
          strcpy(st.comment,*(argv+5));
    } else {

          /* enter any comment */

          printf("Enter any comment -> ");
          read1line(st.comment);      // 1 line read from stdin
    }
    if (argc>=7){
       st.soffset = (long int)atoi(*(argv + 6 ));
    } else {
       st.soffset=0;     // default startoffset
    }

    /* Data header check to get sampling parameters */

    if((kbytes=CheckK5header_with_Read(&K5)) < 0){
        printf("File read error! \n");
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    /* parameter set */

    st.sfreq=K5.sfreq;
    st.adbit=K5.adbit;
    st.numch=K5.numch;
    fsampl=(float)(1000.0*(float)st.sfreq);   // sampling frequency (Hz)
    tsampl=(float)(1.0/fsampl);            // sampling period (sec)
    bits=((long)fsampl)*st.adbit*st.numch;  // # of bits in 1 sec data except header
    bytes=bits/8;                     // # of bytes in 1 sec data except header
    K5.fsample=fsampl;
    //sekibun=1;                       // Averaging period for Spectrum (sec)
    K5.bytesec=bits/8;                  //# of bytes in 1 sec data

    smode=1;                         // low resolution mode

    if(smode ==1){ 
        usampl=10000; 
    } else {
        usampl=50000;
        if(st.sfreq == 40)  usampl=40000; 
        if(st.sfreq == 100) usampl=100000; 
        if(st.sfreq == 200) usampl=200000; 
    }
    imax=(st.sfreq*1000)/usampl;
    numb=usampl*st.adbit*st.numch/8;   // # of bytes in unit # of samples
    K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
    K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second


    tusampl=(double)usampl/K5.fsample;   // period of usampl

    printf("imax numb  %d %d\n",K5.imax,K5.numb);    //debug

    nfft=1024;
    if(usampl < 524288) nfft= 524288;
    if(usampl < 262144) nfft= 262144;
    if(usampl < 131072) nfft= 131072;
    if(usampl < 65536)  nfft= 65536;
    if(usampl < 32768)  nfft= 32768;
    if(usampl < 16384)  nfft= 16384;

    st.nfft=nfft;

    nfact=2.0/(double)nfft;

    K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl
    K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

    K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
    K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

#ifdef USE_FFTW
    cc=fftw_malloc(sizeof(fftw_complex)*nfft);

    pfcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_FORWARD, FFTW_ESTIMATE);
    pbcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif

    cdata = (double *)malloc(2*nfft*sizeof(double));
    st.spows = (float *)malloc(4*nfft*sizeof(float));
    f_hanning = (double *)malloc(nfft*sizeof(double));

    //Hanning filter coeeficients prepare
    for(k=0; k<nfft; k++){
        theta=360.0*(double)k/(double)(nfft-1) - 180.0;
        theta=fmod(theta,360.0) ;
        f_hanning[k]=0.5*(1.0+cos(theta*PIRA));
    }

    fact=((float)usampl)*((float)usampl);

    /* foward startffset seconds */
    if(fwdNsecK5(st.soffset, &K5) <0){
        printf("Foward error %d\n",st.soffset);
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    for(k=0; k<4*nfft; k++) st.spows[k]=0.0;
    for(k=0; k<4; k++) scount[k]=0;

    printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
        K5.hh,K5.mm,K5.ss,K5.seconds);
    sprintf(ltime1st,"Time(UT) %02d:%02d:%02d\0",K5.hh,K5.mm,K5.ss);
    strcpy(st.lab2,ltime1st);
    strcat(st.lab2,"  ");
    strcat(st.lab2,st.comment);
    
        
    tinteg=0.0;

    nloop=0;
loop1:
    if(nloop >0) {
        if((kbytes=CheckK5header_with_Read(&K5)) < 0){
            i=fclose( K5.stream ) ;
            goto exitend;
        }
        printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
            K5.hh,K5.mm,K5.ss,K5.seconds);
    }

    for(i=0; i<K5.imax; i++){
        K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
        K5.pos+=K5.wordread*4;
        K5.vflag=1; // validiate data
        parseADdata_3N(&K5);

        for(m=0; m<st.numch; m++){
            kaisu=0;
            sum=0.0;

            // initialize array
#ifdef USE_FFTW
            for(k=0; k<nfft; k++){ cc[k][0]=0.0; cc[k][1]=0.0;}
#else
            for(k=0; k<2*nfft; k++) cdata[k]=0.0;
#endif

            for(k=0; k<K5.npos/K5.numch; k++){

                /* set data */
#ifdef USE_FFTW
                cc[k][0]=(double)K5.idata[k*st.numch+m];  
                sum+=cc[k][0];
#else
                cdata[2*k]=(double)K5.idata[k*st.numch+m];  
                sum+=cdata[2*k];
#endif
                kaisu++;
            }
             
            dc=sum/(double)kaisu;    // DC component
     
            // DC offset removal from data
            for(k=0; k<K5.npos/st.numch; k++){
#ifdef USE_FFTW
                cc[k][0]-=dc;
#else
                cdata[2*k]-=dc;
#endif
            }

            if(st.amode ==0) {    // spectrum mode
                for(k=0; k<nfft; k++){
#ifdef USE_FFTW
                   cc[k][0]*=f_hanning[k];  // hanning filter
                   cc[k][1]*=f_hanning[k];  // hanning filter
#else
                   cdata[2*k]*=f_hanning[k];
                   cdata[2*k+1]*=f_hanning[k];
#endif
                }
                //hanning(cdata,usampl);  // hanning filter
#ifdef USE_FFTW
                fftw_execute(pbcc);
#else
                cdft(2*nfft,1,cdata);
#endif
            } else {
#ifdef USE_FFTW
                fftw_execute(pfcc);
                for(k=0; k<nfft; k++){
                    cc[k][0]*=nfact; cc[k][1]*=nfact;
                }
#else
                cdft(2*nfft,-1,cdata);
#endif
            }
            for(k=0; k<nfft; k++){
#ifdef USE_FFTW
                pows=(float)((cc[k][0]*cc[k][0]+cc[k][1]*cc[k][1])/(double)fact);
#else
                pows=(float)((cdata[2*k]*cdata[2*k]+cdata[2*k+1]*cdata[2*k+1])/(double)fact);
#endif
                st.spows[k*4+m]=st.spows[k*4+m]+pows;
            }

            scount[m]++;
        }
        tinteg+=tusampl;
        if(tinteg >= st.sekibun) goto exitloop;

    }
    goto loop1;
exitloop:
    i=fclose( K5.stream ) ;

    if(st.amode > 0) {   // auto correlation function calc
       for(m=0; m<st.numch; m++){
          for(k=0; k<nfft; k++){
             st.spows[k*4+m]/=scount[m]; // nomalization
#ifdef USE_FFTW
             cc[k][0]=st.spows[k*4+m]; cc[k][1]=0.0;
#else
             cdata[2*k]=st.spows[k*4+m];
             cdata[2*k+1]=0.0;
#endif
          }
          if(m==1 || m==3){
#ifdef USE_FFTW
              cc[0][0]=1.0;
#else
              cdata[0]=1.0;  // to avoid math error
#endif
          }
#ifdef USE_FFTW
          fftw_execute(pbcc);
#else
          cdft(2*nfft,1,cdata);
#endif
          for(k=0; k<nfft; k++){
             // if(k<10) printf("%d %e\n",k,cdata[2*k]);  // debug
             // if(k>nfft-10) printf("%d %e\n",k,cdata[2*k]);  // debug
#ifdef USE_FFTW
             if(k==0) pmax=(float)cc[0][0];
             if(cc[k][0] > pmax) pmax=(float)cc[k][0];
             st.spows[k*4+m]=(float)cc[k][0];
#else
             if(k==0) pmax=(float)cdata[0];
             if(cdata[2*k] > pmax) pmax=(float)cdata[2*k];
             st.spows[k*4+m]=(float)cdata[2*k];
#endif
          }
          for(k=0; k<nfft; k++){  // normalize
             st.spows[k*4+m]/=pmax;
             // if(k<10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
             // if(k>nfft-10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
          }
             
       }
         
    } else {

        /* convert power to dB  then dBm */       
        cofst=OFFSET1;       //1 bit sampling full power convert to 0 dB
        if( st.adbit==8) cofst=OFFSET8;     // 8bit sampling offset
        if( st.adbit==4) cofst=OFFSET4;     // 4bit sampling offset
        if( st.adbit==2) cofst=OFFSET2;     // 4bit sampling offset
        for(m=0; m<st.numch; m++){
            for(k=0; k<nfft; k++){
                st.spows[k*4+m]/=scount[m]; // nomalization
                st.spows[k*4+m]=(float)(10.0*log10(st.spows[k*4+m]) + cofst); 
                if(k==0){
                    pmaxdbm=st.spows[k*4+m];
                    pmindbm=st.spows[k*4+m];
                    if(m==0){
                        ymaxdb=pmaxdbm;
                        ymindb=pmindbm;
                    }
                } else {
                    if(st.spows[k*4+m] > pmaxdbm) pmaxdbm=st.spows[k*4+m]; 
                    if(st.spows[k*4+m] < pmindbm) pmindbm=st.spows[k*4+m]; 
                }
            }
            pmax=pmaxdbm-cofst;
            pmin=pmindbm-cofst;
            printf("CH# %d: Maximum data (dB, dBm) is %f %f\n", m+1,pmax,pmaxdbm);      
            if(pmaxdbm > ymaxdb) ymaxdb=pmaxdbm;
            if(pmindbm < ymindb) ymindb=pmindbm;
        }
        printf("All CH: Maximum and Minimum data (dBm) are %f %f\n", ymaxdb, ymindb);      

    }
    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);

    /* plot spectrum (or auto cor) POSTSCRIPT OUT and XTERM*/
     plotspe(&st,ymaxdb, ymindb);

  exitend: ;
      
    free(K5.ubuf);
    free(K5.idata);
    free(cdata);
    free(st.spows);

#ifdef USE_FFTW
    fftw_destroy_plan(pfcc);
    fftw_destroy_plan(pbcc);
    fftw_free(cc);
#endif

 exitend2: ;

    return 0;
}
